package com.widget.imagevideobanner.cache;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.os.Build;
import android.os.Environment;
import android.text.TextUtils;
import android.util.Log;
import android.util.LruCache;

import com.widget.imagevideobanner.BuildConfig;
import com.widget.imagevideobanner.cache.disk.DiskLruCache;
import com.widget.imagevideobanner.log.BannerLogFileUtils;
import com.widget.imagevideobanner.utils.BitmapUtils;
import com.widget.imagevideobanner.utils.MD5Util;
import com.widget.imagevideobanner.utils.ThrowableUtils;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import wseemann.media.FFmpegMediaMetadataRetriever;

/**
 * 缓存视频的第一帧图片以及高斯模糊的背景图片
 */
public class BitmapCache {

    static final String TAG = "BitmapCache####";
    static final int BYTES_PER_ARGB_8888_PIXEL = 4;
    //经验值 最多不超过5个视频
    static final int MEMORY_CACHE_TARGET_SCREENS = 10;
    static final int BLUR_RADS = 40;
    static final String BITMAP_CACHE_DIR = Environment.getExternalStorageDirectory().getPath()
            + File.separator
            + "miyaterminal"
            + File.separator
            + "download"
            + File.separator;

    private LruCache<String, Bitmap> memoryCache;
    private DiskLruCache diskLruCache;
    private static BitmapCache instance;
    private ExecutorService executors;
    private Map<String,String> keyMaps = new HashMap<>();

    public static BitmapCache getInstance(Context context) {
        if (instance == null) {
            synchronized (BitmapCache.class) {
                if (instance == null) {
                    instance = new BitmapCache(context);
                }
            }
        }
        return instance;
    }

    public BitmapCache(Context context) {

        int screenSize = context.getResources().getDisplayMetrics().widthPixels
                * context.getResources().getDisplayMetrics().heightPixels
                * BYTES_PER_ARGB_8888_PIXEL;

        //todo 其实应该根据真实的视频个数来比较好
        int targetMemoryCacheSize = Math.round(screenSize * MEMORY_CACHE_TARGET_SCREENS)/1024;
//        int maxMemory = (int) (Runtime.getRuntime().maxMemory() / 1024);
//        int cacheSize = maxMemory / 8;
        //todo 判断是小于最大可用内存
        memoryCache = new LruCache<String, Bitmap>(targetMemoryCacheSize) {
            @Override
            protected int sizeOf(String key, Bitmap bitmap) {
                int size = bitmap.getByteCount() / 1024;
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.KITKAT) {
                    size = bitmap.getAllocationByteCount() / 1024;
                }
                Log.e(TAG,"已经加入内存缓存 size =" +size);
                BannerLogFileUtils.writeLog("已经加入内存缓存 size =" +size);
                return size;
            }
            @Override
            protected void entryRemoved(boolean evicted, String key, Bitmap oldValue, Bitmap newValue) {
                BannerLogFileUtils.writeLog("内存缓存移除 key =" +key);
                Log.e(TAG,"内存缓存移除 key =" +key);
                //todo bitmap复用
                oldValue.recycle();
            }
        };

        try {
            File cacheFile = new File(BITMAP_CACHE_DIR);
            if(!cacheFile.exists() && !cacheFile.mkdir())
               return;
            diskLruCache = DiskLruCache.open(cacheFile, BuildConfig.VERSION_CODE, 1, 100 * 1024 * 1024);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void addBitmapToMemoryCache(String key, Bitmap bitmap) {
        if (getBitmapFromMemCache(key) == null) {
            memoryCache.put(key, bitmap);
        }
    }

    public Bitmap getBitmap(String key){
        if(TextUtils.isEmpty(key = keyMaps.get(key))){
            return null;
        }
        Bitmap bitmap = getBitmapFromMemCache(key);
        if(bitmap != null){
            return bitmap;
        }
        return getBitmapFromDisk(key);
    }

    /**
     * 取高斯模糊的缓存图片
     * @param key
     * @return
     */
    public Bitmap getBlurBitmap(String key){
        if(TextUtils.isEmpty(key = keyMaps.get(key+"blur"))){
            return null;
        }
        Bitmap bitmap = getBitmapFromMemCache(key);
        if(bitmap != null){
            return bitmap;
        }
        return getBitmapFromDisk(key);
    }

    public void addCache(final String url){
        if(TextUtils.isEmpty(url))
            return;

        if(getBitmapFromMemCache(url) != null){
            return;
        }

        if(null == executors)
            executors = Executors.newFixedThreadPool(3);

        executors.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    String key = MD5Util.get32Md5(url);
                    keyMaps.put(url, key);
                    String blurkey = MD5Util.get32Md5(url+"blur");
                    keyMaps.put(url+"blur", blurkey);

                    //截取视频首帧 TODO 横竖屏方向调整
                    FFmpegMediaMetadataRetriever mmr = new FFmpegMediaMetadataRetriever();
                    mmr.setDataSource(url);
                    mmr.extractMetadata(FFmpegMediaMetadataRetriever.METADATA_KEY_ALBUM);
                    mmr.extractMetadata(FFmpegMediaMetadataRetriever.METADATA_KEY_ARTIST);

                    Bitmap b = mmr.getFrameAtTime();
                    //加入到内存MemeryLrucache TODO 采样率压缩
                    addBitmapToMemoryCache(key, b);
                    //加入到DiskLrucache TODO 质量压缩
                    putBitmap2Disk(key, b);

                    //TODO 如果视频比例刚好相等 可以不用背景图片
                    Bitmap blurBitmap = BitmapUtils.createBlurBitmap(b, BLUR_RADS);
                    addBitmapToMemoryCache(blurkey, blurBitmap);
                    putBitmap2Disk(blurkey, blurBitmap);
                }catch (Exception e){
                    e.printStackTrace();
                    BannerLogFileUtils.writeLog("截取视频首帧出错 url ="+url + "出错原因是："+ ThrowableUtils.getFullStackTrace(e));
                }
            }
        });
    }

    /**
     * 从内存缓存获取bitmap
     * @param key
     * @return
     */
    private Bitmap getBitmapFromMemCache(String key) {
        return memoryCache.get(key);
    }


    /**
     * 从磁盘缓存获取bitmap
     * @param key
     * @return
     */
    private Bitmap getBitmapFromDisk(String key) {
        DiskLruCache.Snapshot snapshot = null;
        Bitmap bitmap = null;
        try {
            snapshot = diskLruCache.get(key);
            if (snapshot == null) {
                return null;
            }
            InputStream is = snapshot.getInputStream(0);

            BitmapFactory.Options options = new BitmapFactory.Options();
            bitmap = BitmapFactory.decodeStream(is, null, options);
            if (bitmap != null) {
                memoryCache.put(key, bitmap);
            }

        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            if (snapshot != null) {
                snapshot.close();
            }
        }
        return bitmap;
    }

    /**
     * 放入磁盘缓存
     */
    private void putBitmap2Disk(String key, Bitmap bitmap) {
        DiskLruCache.Snapshot snapshot = null;
        OutputStream os = null;
        try {
            snapshot = diskLruCache.get(key);
            if (snapshot == null) {
                DiskLruCache.Editor edit = diskLruCache.edit(key);
                if (edit != null) {
                    os = edit.newOutputStream(0);
                    bitmap.compress(Bitmap.CompressFormat.JPEG, 50, os);
                    edit.commit();
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            if (snapshot != null) {
                snapshot.close();
            }
            if (os != null) {
                try {
                    os.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }



}
